(* CVT.PAS -- access converted GEOS files
** Copyright (C) 1995,1996  Jochen Metzinger
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

UNIT cvt;

INTERFACE

USES global, fonts;

CONST
 BLOCK_SIZE    = 254; (* size of a data block *)
 BufferMaxSize = $FF*BLOCK_SIZE; (* max. chain length in byte *)

TYPE
 BufferType =
  ARRAY [0..BufferMaxSize-1] OF CHAR;

VAR
 Buffer: ^BufferType;
  (* DMA-Buffer (whole chain), r-o *)
 short_usage: STRING;
  (* of used command *)

PROCEDURE cvt_open(fname: STRING);
(* open .cvt file *)
PROCEDURE cvt_close;
(* close it *)

PROCEDURE cvt_dir(VAR d: DirRec);
(* get directory entry *)
FUNCTION cvt_name: STRING;
(* return file name in directory entry [PC] *)
PROCEDURE cvt_vlir(VAR block: InfoRec);
(* DEBUG: get VLIR record *)
PROCEDURE cvt_info(VAR i: InfoRec);
(* get info record *)
FUNCTION cvt_chain_used(nr: BYTE): BOOLEAN;
(* chain in use? *)
FUNCTION cvt_size(nr: BYTE): LongInt;
(* chain length in byte *)

PROCEDURE cvt_chain(nr: BYTE);
(* reset chain *)
FUNCTION cvt_pos: WORD;
(* get psosition in chain *)
PROCEDURE cvt_seek(pos: WORD);
(* set position in chain *)
FUNCTION cvt_eoc: BOOLEAN;
(* end of chain? *)
PROCEDURE cvt_read(VAR buf; count: WORD; VAR result: WORD);
(* read from chain *)
FUNCTION cvt_getch(VAR ch: CHAR): BOOLEAN;
(* read a single char from chain, return success *)
PROCEDURE cvt_resize(len: LongInt);
(* resize current chain *)

PROCEDURE FATAL(msg: STRING);
(* for FATAL errors *)

IMPLEMENTATION

TYPE
 CVT_BLOCK  = ARRAY [0..BLOCK_SIZE-1] OF BYTE;

VAR
 (* The image file *)
 cvt_file: FILE;
 dir_entry: DirRec;
 info_record: InfoRec;
 vlir: ARRAY [0..126] OF RECORD off, len: LongInt; END;
 (* The chain *)
 chain: BYTE;
 (* The buffer *)
 BufferSize: LongInt;
 BufferPos: WORD;

PROCEDURE cvt_open(fname: STRING);
 VAR block: CVT_BLOCK; res: WORD;
  magic: STRING;
  i: INTEGER; offs: LongInt;
  t, s, s0: BYTE;
BEGIN
 (* open file *)
 {$I-}
 Assign(cvt_file, AddExt(fname, '.'+CVT_EXT));
 Reset(cvt_file,1);
 IF IOResult <> 0 THEN
  FATAL('unable to open file');
 {$I+}
 (* handle .cvt header *)
 BlockRead(cvt_file,block,SizeOf(block),res);
 IF res <> SizeOf(block) THEN
  FATAL('file too short');
 magic := GetString(block[30], 28);
 IF (magic <> 'SEQ formatted GEOS file V1.0')
  AND (magic <> 'PRG formatted GEOS file V1.0') THEN
   FATAL('not a formatted GEOS file V1.0');
 Move(block[0],dir_entry, SizeOf(dir_entry));
 (* handle info record *)
 BlockRead(cvt_file, info_record, SizeOf(info_record),res);
 IF res <> SizeOf(info_record) THEN
  FATAL('file too short');
 (* create vlir[] *)
 IF dir_entry.struct = 0 THEN
  BEGIN (* Sequential structure *)
   WITH vlir[0] DO BEGIN
    off := 2; (* after header/info *)
    len := FileSize(cvt_file) - 2*BLOCK_SIZE;
   END; (* with *)
   FOR i := 1 TO 126 DO
    WITH vlir[i] DO BEGIN
     off := -1;
     len := -1;
    END; (* with *)
  END
 ELSE
  (* VLIR structure *)
  BEGIN
   BlockRead(cvt_file, block, SizeOf(block), res);
   IF res <> SizeOf(block) THEN
    FATAL('file too short');
   offs := 3; (* after header/info/vlir *)
   s0 := 1;
   FOR i := 0 TO 126 DO BEGIN
    t := block[2*i];
    s := block[2*i+1];
    WITH vlir[i] DO
     IF t = 0 THEN
      BEGIN
       off := -1;
       len := -1;
      END
     ELSE
      BEGIN (* used chain *)
       off := offs;
       len := BLOCK_SIZE*(t-1) + (s-1);
       Inc(offs,t);
       s0 := s;
      END; (* with *)
   END; (* for *)
   IF FileSize(cvt_file) < BLOCK_SIZE*(offs-1) + (s0-1) THEN
    FATAL('file too short');
  END; (* else *)
 (* reset other global var. *)
 chain := 0;
 cvt_chain($FF);
END; (* cvt_open *)

PROCEDURE cvt_vlir(VAR block: InfoRec);
 VAR res: WORD;
BEGIN
 IF dir_entry.struct = 0 THEN
  FillChar(block, SizeOf(block), 0)
 ELSE BEGIN
  Seek(cvt_file, 2*BLOCK_SIZE);
  BlockRead(cvt_file, block, SizeOf(block), res);
  IF res <> SizeOf(block) THEN FATAL('READ');
 END; (* else *)
END; (* cvt_vlir *)

PROCEDURE cvt_close;
BEGIN
 Close(cvt_file);
END; (* cvt_close *)

PROCEDURE cvt_dir(VAR d: DirRec);
BEGIN
 d := dir_entry;
END; (* cvt_dir *)

FUNCTION cvt_name: STRING;
 VAR s: STRING; i: INTEGER;
BEGIN
 s := dir_entry.name;
 i := Pos(#$A0, s);
 IF i > 0 THEN Delete(s,i,$FF);
 cvt_name := TransStr(s);
END; (* cvt_name *)

PROCEDURE cvt_info(VAR i: InfoRec);
BEGIN
 i := info_record;
END; (* cvt_info *)

FUNCTION cvt_chain_used(nr: BYTE): BOOLEAN;
BEGIN
 IF nr >= 127 THEN
  cvt_chain_used := FALSE
 ELSE
  cvt_chain_used := (vlir[nr].off > 0);
END; (* cvt_chain_used *)

FUNCTION cvt_size(nr: BYTE): LongInt;
BEGIN
 IF cvt_chain_used(nr)
  THEN cvt_size := vlir[nr].len
  ELSE cvt_size := -1;
END; (* cvt_size *)

PROCEDURE cvt_resize(len: LongInt);
BEGIN
 IF len < 0          THEN len := 0;
 IF len > BufferSize THEN len := BufferSize;
 BufferSize := len;
END; (* cvt_resize *)

PROCEDURE cvt_chain(nr: BYTE);
 VAR result: WORD;
BEGIN
 BufferPos := 0;
 IF chain = nr  THEN EXIT;
 chain := nr;
 BufferSize := 0;
 IF NOT cvt_chain_used(chain) THEN EXIT;
 BufferSize := vlir[chain].len;
 IF BufferSize > BufferMaxSize THEN
  FATAL('chain too long');
 Seek(cvt_file, BLOCK_SIZE*vlir[chain].off);
 BlockRead(cvt_file,Buffer^,BufferSize,result);
 IF result <> BufferSize THEN
  FATAL('READ ERROR');
END; (* cvt_chain *)

FUNCTION cvt_pos: WORD;
BEGIN
 cvt_pos := BufferPos;
END; (* cvt_pos *)

PROCEDURE cvt_seek(pos: WORD);
BEGIN
 IF pos < 0 THEN pos := 0;
 BufferPos := pos;
END; (* cvt_seek *)

PROCEDURE cvt_read(VAR buf; count: WORD; VAR result: WORD);
 VAR MaxCount: LongInt;
BEGIN
 MaxCount := BufferSize - BufferPos;
 IF MaxCount <= 0 THEN
  result := 0
 ELSE BEGIN
  if count > MaxCount THEN
   count := MaxCount;
  Move(Buffer^[BufferPos], buf, count);
  Inc(BufferPos, count);
  result := count;
 END; (* else *)
END; (* cvt_read *)

FUNCTION cvt_eoc: BOOLEAN;
BEGIN
 cvt_eoc := (BufferPos >= BufferSize);
END; (* cvt_eoc *)

FUNCTION cvt_getch(VAR ch: CHAR): BOOLEAN;
BEGIN
 IF NOT cvt_eoc THEN
  BEGIN
   ch := Buffer^[BufferPos];
   Inc(BufferPos);
   cvt_getch := TRUE;
  END
 ELSE
  cvt_getch := FALSE;
END; (* cvt_getch *)

PROCEDURE FATAL(msg: STRING);
BEGIN
 WriteLn(msg);
 IF short_usage <> '' THEN
  WriteLn(short_usage);
 HALT(2);
END; (* FATAL *)

BEGIN
 New(Buffer);
 short_usage := '';
END. (* cvt *)